import pickle
import argparse
from pathlib import Path
from gctree.isotyping import (
    explode_idmap,
    isotype_tree,
    isotype_parsimony,
    default_isotype_order,
)


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=(
            "Given gctree inference outputs, and a file mapping original\n"
            "sequence names (the original sequence ids referenced as values in\n"
            '`idmapfile`) with format "Original SeqID, Isotype", this utility\n\n'
            "* Adds observed isotypes to each observed node in the collapsed\n"
            "  trees output by gctree inference. If cells with the same sequence\n"
            "  but different isotypes are observed, then collapsed tree nodes\n"
            "  must be ‘exploded’ into new nodes with the appropriate isotypes\n"
            "  and abundances. Each unique sequence ID generated by gctree is\n"
            "  prepended to its observed isotype, and a new `isotyped.idmap`\n"
            "  mapping these new sequence IDs to original sequence IDs is \n"
            "  written in the output directory.\n"
            "* Resolves isotypes of unobserved ancestral genotypes in a way\n"
            "  that minimizes isotype switching and obeys isotype switching\n"
            "  order. If observed isotypes of an observed internal node and its\n"
            "  children violate switching order, then the observed internal node\n"
            "  is replaced with an unobserved node with the same sequence, and\n"
            "  the observed internal node is placed as a child leaf. This\n"
            "  procedure always allows switching order conflicts to be resolved,\n"
            "  and should usually increase isotype transitions required in the\n"
            "  resulting tree.\n"
            "* Renders each new collapsed tree with colors and labels\n"
            "  reflecting observed or inferred isotypes, and writes a fasta and\n"
            "  newick file just like the gctree inference pipeline.\n"
            "* Prints for each collapsed tree, the original branching process\n"
            "  log likelihood, the original node count, the isotype parsimony\n"
            "  score, and the new node count after isotype additions. The\n"
            "  isotype parsimony score is just a count of how many isotype\n"
            "  transitions are required along tree edges. Changes in node count\n"
            "  after isotype additions indicate that either observed nodes had\n"
            "  to be exploded based on observed isotypes, or isotype switching\n"
            "  order violations required internal nodes to be expanded as leaf\n"
            "  nodes.\n\n"
            "This tool doesn’t make any judgements about which tree is best.\n"
            "Tree output order is the same as in gctree inference: ranking is\n"
            "by log likelihood before isotype additions. A determination of\n"
            "which is the best tree is left to the user, based on likelihoods,\n"
            "isotype parsimony score, and changes in the number of nodes after\n"
            "isotype additions.\n"
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "idmapfile",
        type=str,
        help="filename for a csv file mapping sequence names to original sequence ids, like the one output by deduplicate.",
    )
    parser.add_argument(
        "isotype_mapfile",
        type=str,
        help="filename for a csv file mapping original sequence ids to observed isotypes"
        ". For example, each line should have the format 'somesequence_id, some_isotype'.",
    )
    parser.add_argument(
        "--trees",
        nargs="+",
        type=str,
        help="filenames for collapsed tree pickle files output by gctree inference",
    )
    parser.add_argument(
        "--isotype_names",
        nargs="+",
        type=str,
        help="A list of isotype names used in isotype_mapfile, in order of most naive to most differentiated."
        """ Default is equivalent to providing the argument ``--isotype_names IgM IgD IgG3 IgG1 IgG2 IgE IgA``""",
    )
    parser.add_argument(
        "--out_directory",
        type=str,
        default=None,
        help="Directory in which to place output. Default is working directory.",
    )
    return parser


def main(arg_list=None):
    isotype_palette = [
        "#a6cee3",
        "#1f78b4",
        "#b2df8a",
        "#33a02c",
        "#fb9a99",
        "#e31a1c",
        "#fdbf6f",
        "#ff7f00",
        "#cab2d6",
        "#6a3d9a",
        "#ffff99",
        "#b15928",
    ]
    args = get_parser().parse_args(arg_list)
    if args.out_directory:
        out_directory = args.out_directory + "/"
        p = Path(out_directory)
        if not p.exists():
            p.mkdir()
    else:
        out_directory = ""
    with open(args.isotype_mapfile, "r") as fh:
        isotypemap = dict(map(lambda x: x.strip(), line.split(",")) for line in fh)

    with open(args.idmapfile, "r") as fh:
        idmap = {}
        for line in fh:
            seqid, cell_ids = line.rstrip().split(",")
            cell_idset = {cell_id for cell_id in cell_ids.split(":") if cell_id}
            if len(cell_idset) > 0:
                idmap[seqid] = cell_idset

    ctrees = []
    for treefile in args.trees:
        with open(treefile, "rb") as fh:
            ctrees.append(pickle.load(fh))
    # parse the idmap file and the isotypemap file
    tree_stats = [
        [
            filename,
            Path(filename).name,
            ctree,
            sum(1 for _ in ctree.tree.traverse()),
        ]
        for filename, ctree in zip(args.trees, ctrees)
    ]
    if not args.isotype_names:
        isotype_names = default_isotype_order
    else:
        isotype_names = args.isotype_names

    newidmap = explode_idmap(idmap, isotypemap)
    for ctree in ctrees:
        ctree.tree = isotype_tree(ctree.tree, newidmap, isotype_names)

    flattened_newidmap = {
        name + " " + str(isotype): cell_idset
        for name, cellid_map in newidmap.items()
        for isotype, cell_idset in cellid_map.items()
    }

    with open(out_directory + "isotyped.idmap", "w") as fh:
        for name, cellidset in flattened_newidmap.items():
            print(f"{name},{':'.join(cellidset)}", file=fh)

    for sublist in tree_stats:
        sublist.append(isotype_parsimony(sublist[2].tree))

    print("name\t\t\t original node count\t isotype parsimony\t new node count")
    for (
        filename,
        name,
        ctree,
        original_numnodes,
        parsimony,
    ) in tree_stats:
        print(
            f"{name}\t\t {original_numnodes}\t\t\t {parsimony}\t\t\t {sum(1 for _ in ctree.tree.traverse())}"
        )
        colormap = {
            node.name: isotype_palette[node.isotype.isotype % len(isotype_palette)]
            for node in ctree.tree.traverse()
        }
        newfilename = name + f".isotype_parsimony.{int(parsimony)}"
        ctree.render(
            outfile=out_directory + newfilename + ".svg",
            colormap=colormap,
            idlabel=True,
        )
        ctree.newick(out_directory + newfilename + ".nk")
        with open(out_directory + newfilename + ".p", "wb") as fh:
            fh.write(pickle.dumps(ctree))



r"""A module for incorporating isotype information in gctree inference."""

from gctree.utils import hamming_distance

import random
import ete3
import warnings
from typing import Dict, Callable, Optional, Set, Sequence, Mapping, Tuple
from functools import wraps
import historydag as hdag
from frozendict import frozendict

default_isotype_order = ["IgM", "IgD", "IgG3", "IgG1", "IgG2", "IgE", "IgA"]


def _assert_switching_order_match(
    fn: Callable[["Isotype", "Isotype"], bool]
) -> Callable:
    @wraps(fn)
    def newfunc(t1: "Isotype", t2: "Isotype") -> bool:
        if t1.order != t2.order:
            raise TypeError(
                "Comparison attempted between isotypes"
                " with different assumed class switching order"
                f" first argument has order {t1.order}"
                f" second argument has order {t2.order}"
            )
        if t1.weight_matrix != t2.weight_matrix:
            raise TypeError(
                "Comparison attempted between isotypes with different weight matrices"
            )
        else:
            return fn(t1, t2)

    return newfunc


class IsotypeTemplate:
    """A template constructor class for :meth:`Isotype`

    Args:
        isotype_order: A list or other Sequence of isotype names, in the allowed class switching order
        weight_matrix: An optional matrix containing transition weights between isotypes. Default weight is 1 for all allowed isotype transitions.
    """

    def __init__(
        self,
        isotype_order: Sequence[str],
        weight_matrix: Optional[Sequence[Sequence[float]]] = None,
    ):
        n = len(isotype_order)
        if weight_matrix is None:
            weight_matrix = [
                [
                    0.0
                    if target == origin
                    else 1.0
                    if target > origin
                    else float("inf")
                    for target in range(n)
                ]
                for origin in range(n)
            ]
        else:
            if len(weight_matrix) != n or len(weight_matrix[0]) != n:
                raise TypeError(
                    "Weight matrix should be an n by n matrix, with n equal to the length of isotype_order."
                )
            # Ensure that illegal transitions are encoded as 'inf' in weight_matrix
            for i in range(n):
                for j in range(n):
                    if i > j:
                        weight_matrix[i][j] = float("inf")
                    elif i == j:
                        weight_matrix[i][j] = 0.0
        self.order = isotype_order
        self.weight_matrix = weight_matrix

    def new(self, isotype_name):
        """Return an :meth:`Isotype` object with the isotype order and weight
        matrix provided to this class's constructor."""
        return Isotype(self.order, self.weight_matrix, isotype_name)


class Isotype:
    """An isotype, and associated information about class switching order and
    transition weights.

    Attributes:
        order: A list or other sequence of isotype names, in their allowed class switching order
        weight_matrix: Transition weights between isotypes, with first index the original isotype index, and second index the new isotype index.
        isotype: Index of isotype name in ``self.order``

    Objects of this class shall be instantiated using :meth:`IsotypeTemplate.new`.
    """

    # From https://doi.org/10.7554/eLife.16578.012, for humans:
    # order = ["M", "G3", "A1", "G2", "G4", "E", "A2"]

    def __init__(
        self,
        order: Sequence[str],
        weight_matrix: Sequence[Sequence[float]],
        isotype_name: str,
    ):
        self.order = order
        self.weight_matrix = weight_matrix
        if isotype_name == "?":
            self.isotype = None
        else:
            try:
                self.isotype = self.order.index(isotype_name)
            except ValueError as e:
                raise ValueError(
                    "Unrecognized isotype name. Check that default"
                    " or provided isotype name list contains all"
                    " observed isotypes\n" + str(e)
                )

    @_assert_switching_order_match
    def isbefore(self, t: "Isotype") -> bool:
        return self.isotype <= t.isotype

    def __repr__(self) -> str:
        return f"Isotype('{self.order[self.isotype]}')"

    def __str__(self) -> str:
        return f"{self.order[self.isotype]}"

    def copy(self) -> "Isotype":
        return Isotype(self.order, self.weight_matrix, self.order[self.isotype])

    @_assert_switching_order_match
    def __eq__(self, other: "Isotype") -> bool:
        return self.isotype == other.isotype

    @_assert_switching_order_match
    def __lt__(self, other: "Isotype") -> bool:
        return self.isotype < other.isotype

    @_assert_switching_order_match
    def __gt__(self, other: "Isotype") -> bool:
        return self.isotype > other.isotype

    def __hash__(self) -> int:
        return hash(self.__repr__())

    def resolutions(self) -> Sequence["Isotype"]:
        """Returns list of all possible isotypes if passed an ambiguous
        isotype.

        Otherwise, returns a list containing a copy of the passed
        isotype.
        """
        if self.isotype is None:
            return [
                Isotype(self.order, self.weight_matrix, name) for name in self.order
            ]
        else:
            return [self.copy()]


def isotype_tree(
    tree: ete3.TreeNode,
    newidmap: Dict[str, Dict[str, str]],
    isotype_names: Sequence[str],
    weight_matrix: Optional[Sequence[Sequence[float]]] = None,
) -> ete3.TreeNode:
    """Method adds isotypes to ``tree``, minimizing isotype switching and
    obeying switching order.

    * Adds observed isotypes to each observed node in the collapsed
      trees output by gctree inference. If cells with the same sequence
      but different isotypes are observed, then collapsed tree nodes
      must be ‘exploded’ into new nodes with the appropriate isotypes
      and abundances. Each unique sequence ID generated by gctree is
      prepended to its observed isotype, and a new `isotyped.idmap`
      mapping these new sequence IDs to original sequence IDs is
      written in the output directory.
    * Resolves isotypes of unobserved ancestral genotypes in a way
      that minimizes isotype switching and obeys isotype switching
      order. If observed isotypes of an observed internal node and its
      children violate switching order, then the observed internal node
      is replaced with an unobserved node with the same sequence, and
      the observed internal node is placed as a child leaf. This
      procedure always allows switching order conflicts to be resolved,
      and should usually increase isotype transitions required in the
      resulting tree.

    Args:
        tree: ete3 Tree
        newidmap: mapping of sequence IDs to isotypes, such as that output by :meth:`utils.explode_idmap`.
        isotype_names: list or other sequence of isotype names observed, in correct switching order.

    Returns:
        A new ete3 Tree whose nodes have isotype annotations in the attribute ``isotype``.
        Node names in this tree also contain isotype names.
    """
    tree = tree.copy()
    _add_observed_isotypes(tree, newidmap, isotype_names, weight_matrix=weight_matrix)
    _disambiguate_isotype(tree)
    _collapse_tree_by_sequence_and_isotype(tree)
    for node in tree.traverse():
        node.name = str(node.name) + " " + str(node.isotype)
    for node in tree.iter_descendants():
        node.dist = hamming_distance(node.up.sequence, node.sequence)
    return tree


@_assert_switching_order_match
def isotype_distance(t1: Isotype, t2: Isotype) -> float:
    """Returns the weight of the transition from isotype ``t1`` to isotype
    ``t2``.

    This function is not symmetric on its arguments, so isn't a true
    distance.
    """
    return t1.weight_matrix[t1.isotype][t2.isotype]


def isotype_parsimony(tree: ete3.TreeNode) -> float:
    """Computes the sum of :meth:`isotype_distance` along each edge in an
    isotyped tree.

    If no weight matrix was provided during isotyping of the tree, then
    the return value of this function is the number of isotype
    transitions along edges in the tree.
    """
    return sum(
        isotype_distance(node.up.isotype, node.isotype)
        for node in tree.iter_descendants()
    )


def _disambiguate_isotype(
    tree: ete3.Tree, random_state=None, dist_func=isotype_distance
):
    def is_ambiguous(isotype):
        return isotype.isotype is None

    if random_state is None:
        random.seed(tree.write(format=1))
    else:
        random.setstate(random_state)

    # First pass of Sankoff: compute cost vectors
    for node2 in tree.traverse(strategy="postorder"):
        node2.add_feature("costs", [[seq, 0] for seq in node2.isotype.resolutions()])
        if not node2.is_leaf():
            for seq_cost in node2.costs:
                for child in node2.children:
                    seq_cost[1] += min(
                        [
                            dist_func(seq_cost[0], child_seq) + child_cost
                            for child_seq, child_cost in child.costs
                        ]
                    )
    # Second pass: Choose base and adjust children's cost vectors.
    # Not necessary if we only want the minimum weight:
    for node2 in tree.traverse(strategy="preorder"):
        if not is_ambiguous(node2.isotype):
            continue
        min_cost = min([cost for _, cost in node2.costs])
        resolved_isotype = random.choice(
            [isotype for isotype, cost in node2.costs if cost == min_cost]
        )
        # Adjust child cost vectors
        if not node2.is_leaf():
            for child in node2.children:
                child.costs = [
                    [
                        isotype,
                        cost + dist_func(resolved_isotype, isotype),
                    ]
                    for isotype, cost in child.costs
                ]
        node2.isotype = resolved_isotype
    for node in tree.traverse():
        try:
            node.del_feature("costs")
        except (AttributeError, KeyError):
            pass


def _add_observed_isotypes(
    tree: ete3.Tree,
    newidmap: Dict[str, str],
    isotype_order: Sequence[str],
    weight_matrix: Optional[Sequence[Sequence[float]]] = None,
):
    # Drop observed nodes as leaves and explode by observed isotype:
    # Descend internal observed nodes as leaves:
    newisotype = IsotypeTemplate(isotype_order, weight_matrix=weight_matrix).new
    for node in list(tree.iter_descendants()):
        if node.abundance > 0 and not node.is_leaf():
            newchild = ete3.TreeNode(name=node.name)
            newchild.add_feature("sequence", node.sequence)
            newchild.add_feature("abundance", node.abundance)
            node.abundance = 0
            node.add_child(child=newchild)
    # Now duplicate nodes which represent multiple isotypes
    for node in list(tree.get_leaves()):
        if node.abundance == 0:
            node.add_feature("isotype", newisotype("?"))
        else:
            try:
                thisnode_isotypemap = newidmap[node.name]
            except KeyError as e:
                warnings.warn(
                    f"The sequence name {e} labels an observed node, but no mapping to an original sequence ID was found."
                    " Isotype will be assumed ambiguous."
                )
                thisnode_isotypemap = {
                    "?": {f"Unknown_id_{n+1}" for n in range(node.abundance)}
                }
            if "?" in thisnode_isotypemap:
                warnings.warn(
                    f"The sequence name {node.name} labels an observed node, and corresponds to sequence IDs for "
                    "which no observed isotype was provided. "
                    f" Isotype will be assumed ambiguous for: {', '.join(thisnode_isotypemap['?'])}"
                )
            # node.name had better be in newidmap, since this is an observed node
            if len(thisnode_isotypemap) > 1:
                for isotype, cell_ids in thisnode_isotypemap.items():
                    # add new node below this leaf node. Must be below, and not child
                    # of parent, to preserve max parsimony in case that node.up has
                    # different sequence from node.
                    newchild = ete3.TreeNode(name=node.name)
                    newchild.add_feature("abundance", len(cell_ids))
                    newchild.add_feature("sequence", node.sequence)
                    newchild.add_feature("isotype", newisotype(isotype))
                    node.add_child(child=newchild)
                node.abundance = 0
            else:
                node.isotype = newisotype(list(thisnode_isotypemap.keys())[0])
    # Now add ancestral ambiguous isotypes
    for node in tree.traverse():
        if not node.is_leaf():
            node.add_feature("isotype", newisotype("?"))


def _collapse_tree_by_sequence_and_isotype(tree: ete3.TreeNode):
    for node in tree.iter_descendants():
        node.dist = node.up.sequence != node.sequence or node.up.isotype != node.isotype
    for node in tree.iter_descendants():
        if node.dist == 0:
            node.up.abundance += node.abundance
            node.up.name = node.name
            node.delete(prevent_nondicotomic=False)


def explode_idmap(
    idmap: Dict[str, Set[str]], isotype_map: Dict[str, str]
) -> Dict[str, Dict[str, Set[str]]]:
    """Uses provided mapping of original sequence IDs to observed isotypes to
    'explode' the provided mapping of unique sequence IDs to original sequence
    IDs.

    Args:
        idmap: A dictionary mapping unique sequence IDs to sets of original IDs of observed sequences
        isotype_map: A dictionary mapping original IDs to observed isotype names

    Returns:
        A dictionary mapping unique sequence IDs to dictionaries, mapping isotype names to sets of original sequence IDs.
    """
    newidmap = {}
    for id, cell_ids in idmap.items():
        isotypeset = set()
        for cell_id in cell_ids:
            try:
                isotypeset.add(isotype_map[cell_id])
            except KeyError:
                warnings.warn(
                    f"Sequence ID {id} has original sequence id {cell_id} "
                    "for which no observed isotype was provided. "
                    "Isotype will be assumed ambiguous if observed."
                )
                isotype_map[cell_id] = "?"
                isotypeset.add("?")
        newidmap[id] = {
            isotype: {
                cell_id for cell_id in cell_ids if isotype_map[cell_id] == isotype
            }
            for isotype in isotypeset
        }
    return newidmap


def _isotype_dagfuncs() -> hdag.utils.AddFuncDict:
    """Return functions for filtering by isotype parsimony score on the history
    DAG.

    The DAG on which these functions is run should have key ``isotype`` in their attr dictionaries.
    The :meth:``CollapsedForest.add_isotypes`` method can be used to achieve this.
    If isotypes aren't added, these functions will return a weight of 0 for all trees.

    Isotype parsimony of a tree is the minimum number of allowed isotype transitions
    required along all edges.

    Returns:
        A :meth:`historydag.utils.AddFuncDict` which may be passed as keyword arguments
        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
        methods to trim or annotate a :meth:`historydag.HistoryDag` according to isotype parsimony.
        Weight type is ``int``.
    """

    def edge_weight_func(n1: hdag.HistoryDagNode, n2: hdag.HistoryDagNode):
        if n1.is_root():
            return 0
        else:
            n1isos = n1.attr["isotype"]
            n2isos = n2.attr["isotype"]
            if not n1isos or not n2isos:
                return 0
            n1iso = list(n1isos.keys())[0]
            return int(sum(isotype_distance(n1iso, n2iso) for n2iso in n2isos.keys()))

    return hdag.utils.AddFuncDict(
        {
            "start_func": lambda n: 0,
            "edge_weight_func": edge_weight_func,
            "accum_func": sum,
        },
        name="Isotype Pars.",
    )


def _isotype_annotation_dagfuncs(
    isotypemap: Mapping[str, str] = None,
    isotypemap_file: str = None,
    idmap: Mapping[str, Set[str]] = None,
    idmap_file: str = None,
    isotype_names: Sequence[str] = None,
) -> hdag.utils.AddFuncDict:
    """Return functions for annotating history DAG nodes with inferred
    isotypes.

    Args:
        isotypemap: A dictionary mapping original IDs to observed isotype names
        isotypemap_file: A csv file providing an `isotypemap`
        idmap: A dictionary mapping unique sequence IDs to sets of original IDs of observed sequences
        idmap_file: A csv file providing an `idmap`
        isotype_names: A sequence of isotype names containing values in `isotypemap`, in the correct switching order

    Returns:
        A :meth:`historydag.utils.AddFuncDict` which may be passed as keyword arguments
        to :meth:`historydag.HistoryDag.weight_count`, :meth:`historydag.HistoryDag.trim_optimal_weight`,
        or :meth:`historydag.HistoryDag.optimal_weight_annotate`
        methods to trim or annotate a :meth:`historydag.HistoryDag` according to isotype parsimony.
        Weight format is ``frozendict[Isotype, int]``, where the value is the count of observed isotypes, and 0
        for unobserved (internal) nodes.
    """
    if isotype_names is None:
        isotype_names = default_isotype_order
    if isotypemap_file and isotypemap is None:
        with open(isotypemap_file, "r") as fh:
            isotypemap = dict(map(lambda x: x.strip(), line.split(",")) for line in fh)
    elif isotypemap_file is None:
        raise ValueError("Either isotypemap or isotypemap_file is required")

    if idmap is None:
        if idmap_file is None:
            raise TypeError("either idmap or idmap_file is required for isotyping")
        else:
            with open(idmap_file, "r") as fh:
                idmap = {}
                for line in fh:
                    seqid, cell_ids = line.rstrip().split(",")
                    cell_idset = {cell_id for cell_id in cell_ids.split(":") if cell_id}
                    if len(cell_idset) > 0:
                        idmap[seqid] = cell_idset
    newidmap = explode_idmap(idmap, isotypemap)
    newisotype = IsotypeTemplate(isotype_names, weight_matrix=None).new

    def start_func(n: hdag.HistoryDagNode):
        seqid = n.attr["name"]
        if seqid in newidmap:
            return frozendict(
                {
                    newisotype(name): len(oldidset)
                    for name, oldidset in newidmap[seqid].items()
                }
            )
        else:
            return frozendict()

    def sumweights(weights: Sequence[Tuple[float, frozendict]]):
        ts = [item for i in weights for item in i.keys()]
        if ts:
            return frozendict({min(ts): 0})
        else:
            return frozendict()

    return hdag.utils.AddFuncDict(
        {
            "start_func": start_func,
            "edge_weight_func": lambda n1, n2: frozendict(),
            "accum_func": sumweights,
        },
        name="Inferred Isotype",
    )



#! /usr/bin/env python

import gctree.branching_processes as bp
import gctree.phylip_parse as pp
import gctree.mutation_model as mm
import gctree.utils as utils


import argparse
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import warnings
import pickle
import random
import ete3
import itertools


def test(args):
    """Test subprogram.

    Checks likelihood against a by-hand calculation for a simple tree,
    simulates a forest, computes MLE parameters, and plots some sanity
    check figures to plot_file command line arguments are p, q, number
    of trees to simulate, and plot file name
    """

    import seaborn as sns

    sns.set(style="white", color_codes=True)
    sns.set_style("ticks")

    # compare likelihood to empirical likelihood (might need large n)
    n = 10000
    df = pd.DataFrame(columns=("p", "q", "parameters", "f", "L"))
    ct = 0
    ps = (0.1, 0.2, 0.3, 0.4)
    qs = (0.2, 0.4, 0.6, 0.8)
    np.seterr(all="ignore")
    for p in ps:
        for q in qs:
            forest = bp.CollapsedForest()
            if args.verbose:
                print(f"parameters: p = {p}, q = {q}")
            forest.simulate(p, q, n)
            tree_dict = {}
            for tree in forest:
                tree_hash = tuple(
                    (node.abundance, len(node.children))
                    for node in tree.tree.traverse()
                )
                if tree_hash not in tree_dict:
                    tree_dict[tree_hash] = [tree, tree.ll(p, q)[0], 1]
                else:
                    tree_dict[tree_hash][-1] += 1
            L_empirical, L_theoretical = zip(
                *[
                    (tree_dict[tree_hash][2], np.exp(tree_dict[tree_hash][1]))
                    for tree_hash in tree_dict
                    if tree_dict[tree_hash][2]
                ]
            )
            for tree_hash in tree_dict:
                df.loc[ct] = (
                    p,
                    q,
                    f"p={p}, q={q}",
                    tree_dict[tree_hash][2],
                    np.exp(tree_dict[tree_hash][1]),
                )
                ct += 1
    if args.verbose:
        print()

    plt.figure()
    limx = (1 / n, 1.1)
    limy = (1, 1.1 * n)
    g = sns.lmplot(
        x="L",
        y="f",
        col="p",
        row="q",
        hue="parameters",
        data=df,
        fit_reg=False,
        scatter_kws={"alpha": 0.3},
        height=1.5,
        legend=False,
        row_order=reversed(qs),
    )
    g.set(xscale="log", yscale="log", xlim=limx, ylim=limy)
    for i in range(len(ps)):
        for j in range(len(qs)):
            g.axes[i, j].plot(
                limx, limy, ls="--", c="black", lw=0.5, zorder=0, markeredgewidth=0.1
            )
            g.axes[i, j].set_title(
                f"p={ps[j]}\nq={list(reversed(qs))[i]}",
                x=0.05,
                y=0.7,
                size="x-small",
                ha="left",
            )
    g.set_axis_labels("", "")
    g.fig.text(0.45, 0.02, s="gctree likelihood", multialignment="center")
    g.fig.text(
        0.02,
        0.7,
        s=f"frequency among {n} simulations",
        rotation=90,
        multialignment="center",
    )
    plt.savefig(args.outbase + "." + args.img_type)

    # MLE check
    n = 20
    n2 = 500
    df = pd.DataFrame(columns=("true parameters", "$\\hat{p}$", "$\\hat{q}$"))
    ct = 0
    for p in ps:
        for q in qs:
            for _ in range(n):
                forest = bp.CollapsedForest()
                if args.verbose:
                    print(f"parameters: p = {p}, q = {q}")
                forest.simulate(p, q, n2)
                p_inferred, q_inferred = forest.mle()
                df.loc[ct] = (f"p={p}, q={q}", p_inferred, q_inferred)
                ct += 1

    plt.figure()
    g = sns.lmplot(
        x="$\\hat{p}$",
        y="$\\hat{q}$",
        hue="true parameters",
        data=df,
        fit_reg=False,
        scatter_kws={"alpha": 0.2},
        height=6,
        legend=False,
    )
    g.set(
        xlim=(0.05, 0.45),
        xticks=np.arange(0.0, 0.6, 0.1),
        ylim=(0.1, 0.9),
        yticks=np.arange(0.0, 1.2, 0.2),
    )
    for i in range(len(ps)):
        for j in range(len(qs)):
            plt.scatter([ps[i]], [qs[j]], c="black", marker="+")
    plt.savefig(args.outbase + ".2." + args.img_type)

    return


def infer(args):
    """inference subprogram."""

    def isotype_add(forest):
        forest.add_isotypes(
            isotypemap_file=args.isotype_mapfile,
            idmap_file=args.idmapfile,
            isotype_names=args.isotype_names,
        )

    if len(args.infiles) == 2:
        forest = bp.CollapsedForest(
            pp.parse_outfile(args.infiles[0], args.infiles[1], args.root)
        )
        if forest.n_trees == 1:
            warnings.warn("only one parsimony tree reported from dnapars")

        if args.verbose:
            print("number of trees with integer branch lengths:", forest.n_trees)
        forest.mle(marginal=True)
        # Add isotypes to forest
        if args.isotype_mapfile:
            isotype_add(forest)
        with open(args.outbase + ".inference.parsimony_forest.p", "wb") as f:
            pickle.dump(forest, f)

    elif len(args.infiles) == 1:
        if args.verbose:
            print(
                "Loading provided parsimony forest. If forest has fit parameters, parameter fitting will be skipped."
            )
        with open(args.infiles[0], "rb") as fh:
            forest = pickle.load(fh)
        # Add isotypes to forest and re-pickle if necessary
        if args.isotype_mapfile:
            isotype_add(forest)
            with open(args.outbase + ".inference.parsimony_forest.p", "wb") as f:
                pickle.dump(forest, f)
    else:
        raise ValueError(
            "The filename of a pickled history DAG object, or a phylipfile and abundance file, are required."
        )

    # Filter the forest according to specified criteria, and along the way,
    # write a log file containing stats for all trees in the forest:
    trimmed_forest, _ = forest.filter_trees(
        ranking_coeffs=args.ranking_coeffs,
        verbose=args.verbose,
        outbase=args.outbase,
        summarize_forest=args.summarize_forest,
        tree_stats=args.tree_stats,
        mutability_file=args.mutability,
        substitution_file=args.substitution,
        chain_split=args.chain_split,
    )

    if args.verbose:
        if trimmed_forest.n_trees > 1:
            n_topologies = trimmed_forest.n_topologies()
            print(
                "Degenerate ranking criteria: trimmed history DAG contains "
                f"{trimmed_forest.n_trees} unique trees, with {n_topologies} unique collapsed topologies."
            )
            if n_topologies > 10:
                print(
                    "A representative from each of the ten most abundant topologies will be sampled randomly for rendering."
                )
            else:
                print(
                    "A representative of each topology will be sampled randomly for rendering."
                )

    random.seed(forest.n_trees)
    ctrees = [
        topoclass.sample_tree()
        for _, topoclass in zip(range(10), trimmed_forest.iter_topology_classes())
    ]

    if args.colormapfile is not None:
        with open(args.colormapfile, "r") as f:
            colormap = {}
            for line in f:
                seqid, color = line.rstrip().split("\t")
                if "," in color:
                    colors = {
                        x.split(":")[0]: int(x.split(":")[1]) for x in color.split(",")
                    }
                    colormap[seqid] = colors
                else:
                    colormap[seqid] = color
    else:
        colormap = None

    # parse position map file(s)
    if args.positionmapfile is not None:
        with open(args.positionmapfile) as f:
            position_map = [int(x) for x in f.read().split()]
    else:
        position_map = None

    if args.positionmapfile2 is not None:
        with open(args.positionmapfile2) as f:
            position_map2 = [int(x) for x in f.read().split()]
    else:
        position_map2 = None

    for j, collapsed_tree in enumerate(ctrees, 1):
        collapsed_tree.render(
            f"{args.outbase}.inference.{j}.svg",
            idlabel=args.idlabel,
            colormap=colormap,
            frame=args.frame,
            position_map=position_map,
            chain_split=args.chain_split,
            frame2=args.frame2,
            position_map2=position_map2,
        )
        collapsed_tree.newick(f"{args.outbase}.inference.{j}.nk")
        with open(f"{args.outbase}.inference.{j}.p", "wb") as f:
            pickle.dump(collapsed_tree, f)

    # rank plot of observed allele frequencies
    y = sorted(
        (node.abundance for node in ctrees[0].tree.traverse() if node.abundance != 0),
        reverse=True,
    )
    plt.figure()
    plt.bar(range(1, len(y) + 1), y, color="black")
    plt.xlabel("genotype")
    plt.ylabel("abundance")
    plt.savefig(args.outbase + ".inference.abundance_rank." + args.img_type)


def simulate(args):
    """Simulation subprogram.

    Simulates a Galton–Watson process, with mutation probabilities
    according to a user defined motif model e.g. S5F
    """
    random.seed(a=args.seed)
    mutation_model = mm.MutationModel(args.mutability, args.substitution)
    if args.lambda0 is None:
        args.lambda0 = [max([1, int(0.01 * len(args.sequence))])]
    args.sequence = args.sequence.upper()
    if args.sequence2 is not None:
        # Use the same mutation rate on both sequences
        if len(args.lambda0) == 1:
            args.lambda0 = [args.lambda0[0], args.lambda0[0]]
        elif len(args.lambda0) != 2:
            raise Exception(
                "Only one or two lambda0 can be defined for a two "
                "sequence simulation."
            )
        # Require both sequences to be in frame 1:
        if args.frame is not None and args.frame != 1:
            if args.verbose:
                print(
                    "Warning: When simulating with two sequences they are "
                    "truncated to be beginning at frame 1."
                )
            args.sequence = args.sequence[
                (args.frame - 1) : (
                    args.frame
                    - 1
                    + (3 * (((len(args.sequence) - (args.frame - 1)) // 3)))
                )
            ]
            args.sequence2 = args.sequence2[
                (args.frame - 1) : (
                    args.frame
                    - 1
                    + (3 * (((len(args.sequence2) - (args.frame - 1)) // 3)))
                )
            ]
        # Extract the bounds between sequence 1 and 2:
        seq_bounds = (
            (0, len(args.sequence)),
            (len(args.sequence), len(args.sequence) + len(args.sequence2)),
        )
        # Merge the two seqeunces to simplify future dealing with the pair:
        args.sequence += args.sequence2
    else:
        seq_bounds = None

    trials = 1000
    # this loop makes us resimulate if size too small, or backmutation
    for trial in range(trials):
        try:
            tree = mutation_model.simulate(
                args.sequence,
                seq_bounds=seq_bounds,
                progeny=lambda seq: args.lambda_,
                lambda0=args.lambda0,
                n=args.n,
                N=args.N,
                T=args.T,
                frame=args.frame,
                verbose=args.verbose,
            )
            # this will fail if backmutations
            collapsed_tree = bp.CollapsedTree(tree=tree)
            tree.ladderize()
            uniques = sum(node.abundance > 0 for node in collapsed_tree.tree.traverse())
            if uniques < 2:
                raise RuntimeError(
                    f"collapsed tree contains {uniques} " "sampled sequences"
                )
            break
        except RuntimeError as e:
            print(f"{e}, trying again")
        else:
            raise
    if trial == trials - 1:
        raise RuntimeError(f"{trials} attempts exceeded")

    # In the case of a sequence pair print them to separate files:
    if args.sequence2 is not None:
        fh1 = open(args.outbase + ".simulation_seq1.fasta", "w")
        fh2 = open(args.outbase + ".simulation_seq2.fasta", "w")
        fh1.write(">root\n")
        fh1.write(args.sequence[seq_bounds[0][0] : seq_bounds[0][1]] + "\n")
        fh2.write(">root\n")
        fh2.write(args.sequence[seq_bounds[1][0] : seq_bounds[1][1]] + "\n")
        for leaf in tree.iter_leaves():
            if leaf.abundance != 0:
                fh1.write(">" + leaf.name + "\n")
                fh1.write(leaf.sequence[seq_bounds[0][0] : seq_bounds[0][1]] + "\n")
                fh2.write(">" + leaf.name + "\n")
                fh2.write(leaf.sequence[seq_bounds[1][0] : seq_bounds[1][1]] + "\n")
    else:
        with open(args.outbase + ".simulation.fasta", "w") as f:
            f.write(">root\n")
            f.write(args.sequence + "\n")
            for leaf in tree.iter_leaves():
                if leaf.abundance != 0:
                    f.write(">" + leaf.name + "\n")
                    f.write(leaf.sequence + "\n")

    # some observable simulation stats to write
    abundance, distance_from_root, degree = zip(
        *[
            (
                node.abundance,
                utils.hamming_distance(node.sequence, args.sequence),
                sum(
                    utils.hamming_distance(node.sequence, node2.sequence) == 1
                    for node2 in collapsed_tree.tree.traverse()
                    if node2.abundance and node2 is not node
                ),
            )
            for node in collapsed_tree.tree.traverse()
            if node.abundance
        ]
    )
    stats = pd.DataFrame(
        {
            "genotype abundance": abundance,
            "Hamming distance to root genotype": distance_from_root,
            "Hamming neighbor genotypes": degree,
        }
    )
    stats.to_csv(args.outbase + ".simulation.stats.tsv", sep="\t", index=False)

    print(
        f"{sum(leaf.abundance for leaf in collapsed_tree.tree.traverse())}"
        " simulated observed sequences"
    )

    # render the full lineage tree
    ts = ete3.TreeStyle()
    ts.rotation = 90
    ts.show_leaf_name = False
    ts.show_scale = False

    colors = {}
    palette = ete3.SVG_COLORS
    palette -= set(["black", "white", "gray"])
    palette = itertools.cycle(list(palette))  # <-- circular iterator

    colors[tree.sequence] = "gray"

    for n in tree.traverse():
        nstyle = ete3.NodeStyle()
        nstyle["size"] = 10
        if args.plotAA:
            if n.AAseq not in colors:
                colors[n.AAseq] = next(palette)
            nstyle["fgcolor"] = colors[n.AAseq]
        else:
            if n.sequence not in colors:
                colors[n.sequence] = next(palette)
            nstyle["fgcolor"] = colors[n.sequence]
        n.set_style(nstyle)

    # this makes the rendered branch lenths correspond to time
    for node in tree.iter_descendants():
        node.dist = node.time - node.up.time
    tree.render(args.outbase + ".simulation.lineage_tree.svg", tree_style=ts)

    # render collapsed tree
    # create an id-wise colormap
    # NOTE: node.name can be a set
    colormap = {
        node.name: colors[node.sequence] for node in collapsed_tree.tree.traverse()
    }
    collapsed_tree.write(args.outbase + ".simulation.collapsed_tree.p")
    collapsed_tree.render(
        args.outbase + ".simulation.collapsed_tree.svg",
        idlabel=args.idlabel,
        colormap=colormap,
        frame=args.frame,
    )
    # print colormap to file
    with open(args.outbase + ".simulation.collapsed_tree.colormap.tsv", "w") as f:
        for name, color in colormap.items():
            f.write(
                (name if isinstance(name, str) else ",".join(name))
                + "\t"
                + color
                + "\n"
            )


def get_parser():
    parser = argparse.ArgumentParser(
        description="genotype collapsed tree inference and simulation"
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        description="specify one of these",
        required=True,
        help="additional help available for each subcommand",
    )

    # parser for test subprogram
    parser_test = subparsers.add_parser(
        "test",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="run tests on library functions",
    )
    parser_test.set_defaults(func=test)

    # parser for inference subprogram
    parser_infer = subparsers.add_parser(
        "infer",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="likelihood ranking of parsimony trees",
    )
    parser_infer.add_argument(
        "--root",
        type=str,
        default="root",
        help=r'name of root sequence (outgroup root), default ``"root"``',
    )
    parser_infer.add_argument(
        "infiles",
        type=str,
        nargs="+",
        help=(
            "Input files for inference. If two filenames are passed, the first shall be a "
            "dnapars outfile (verbose output with sequences at each site), and the second "
            "shall be an abundance file containing allele frequencies (sequence counts) in "
            "the format: ``SeqID, Nobs``. If a single filename is passed, it shall be the name "
            "of a pickled history DAG object created by gctree. A new pickled forest will be "
            "output only if new annotations (such as isotypes) are added."
        ),
    )
    parser_infer.add_argument(
        "--colormapfile",
        type=str,
        default=None,
        help='File containing color map in the tab-separated format: ``"SeqID\tcolor"``',
    )
    parser_infer.add_argument(
        "--chain_split",
        type=int,
        default=None,
        help=(
            "when using concatenated heavy and light chains, this is the 0-based"
            " index at which the 2nd chain begins, needed for determining coding frame in both chains,"
            " and also to correctly calculate mutability parsimony."
        ),
    )
    parser_infer.add_argument(
        "--frame2",
        type=int,
        default=None,
        choices=(1, 2, 3),
        help=r"codon frame for the second chain when using the ``chain_split`` option",
    )
    parser_infer.add_argument(
        "--positionmapfile",
        type=str,
        default=None,
        help="file containing a list of position numbers (e.g. IMGT) corresponding to indices in sequence",
    )
    parser_infer.add_argument(
        "--positionmapfile2",
        type=str,
        default=None,
        help="positionmapfile for the 2nd chain when using the ``chain_split`` option",
    )
    parser_infer.set_defaults(func=infer)
    parser_infer.add_argument(
        "--idmapfile",
        default=None,
        type=str,
        help=(
            "input filename for a csv file mapping sequence names to original sequence ids. "
            "For use by isotype ranking. "
            "Such a file can be produced by ``deduplicate`` when it is provided the ``--idmapfile`` option. "
        ),
    )
    parser_infer.add_argument(
        "--isotype_mapfile",
        default=None,
        type=str,
        help=(
            "filename for a csv file mapping original sequence ids to observed isotypes"
            ". For example, each line should have the format 'somesequence_id, some_isotype'."
        ),
    )
    parser_infer.add_argument(
        "--isotype_names",
        type=str,
        default=None,
        nargs="+",
        help=(
            "A list of isotype names used in isotype_mapfile, in order of most naive to most differentiated."
            """ Default is equivalent to providing the argument ``--isotype_names IgM IgD IgG3 IgG1 IgG2 IgE IgA``"""
        ),
    )
    parser_infer.add_argument(
        "--mutability",
        type=str,
        default=None,
        help=(
            "Path to mutability model file. If --mutability and --substitution are both provided, "
            "they will be used to rank trees after likelihood and isotype parsimony."
        ),
    )
    parser_infer.add_argument(
        "--substitution",
        type=str,
        default=None,
        help=(
            "Path to substitution model file. If --mutability and --substitution are both provided, "
            "they will be used to rank trees after likelihood and isotype parsimony."
        ),
    )
    parser_infer.add_argument(
        "--ranking_coeffs",
        type=float,
        nargs=3,
        default=None,
        help=(
            "List of coefficients for ranking trees by a linear combination of traits. "
            "Coefficients are in order: isotype parsimony, mutation model parsimony, number of alleles. "
            "A coefficient of -1 will be applied to branching process likelihood. "
            "If not provided, trees will be ranked lexicographically by likelihood, "
            "isotype parsimony, and mutability parsimony in that order."
        ),
    )
    parser_infer.add_argument(
        "--summarize_forest",
        action="store_true",
        help=(
            "write a file `[outbase].forest_summary.log` with a summary of traits for trees in the forest."
        ),
    )
    parser_infer.add_argument(
        "--tree_stats",
        action="store_true",
        help=(
            "write a file `[outbase].tree_stats.log` with stats for all trees in the forest. "
            "For large forests, this is slow and memory intensive."
        ),
    )

    # parser for simulation subprogram
    parser_sim = subparsers.add_parser(
        "simulate",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Neutral, and target selective, model gctree simulation",
    )
    parser_sim.add_argument("sequence", type=str, help="seed root nucleotide sequence")
    parser_sim.add_argument(
        "mutability", type=str, help="path to mutability model file"
    )
    parser_sim.add_argument(
        "substitution", type=str, help="path to substitution model file"
    )
    parser_sim.add_argument(
        "--sequence2",
        type=str,
        default=None,
        help="Second seed root nucleotide sequence. "
        "For simulating heavy/light chain co-evolution.",
    )
    parser_sim.add_argument(
        "--lambda",
        dest="lambda_",
        type=float,
        default=0.9,
        help="poisson branching parameter",
    )
    parser_sim.add_argument(
        "--lambda0",
        type=float,
        default=None,
        nargs="*",
        help="List of one or two elements with the baseline mutation rates. "
        "Space separated input values. First element belonging to seed "
        "sequence one and optionally the next to sequence 2. If only one "
        "rate is provided for two sequences, this rate will be used on "
        "both.",
    )
    parser_sim.add_argument("--n", type=int, default=None, help="cells downsampled")
    parser_sim.add_argument(
        "--N", type=int, default=None, help="target simulation size"
    )
    parser_sim.add_argument(
        "--T",
        type=int,
        nargs="+",
        default=None,
        help="observation time, if None we run until termination and take all"
        " leaves",
    )
    parser_sim.add_argument(
        "--seed", type=int, default=None, help="integer random seed"
    )
    parser_sim.add_argument(
        "--target_dist",
        type=int,
        default=10,
        help="The number of non-synonymous mutations the target should be "
        "away from the root.",
    )
    parser_sim.add_argument(
        "--plotAA",
        type=bool,
        default=False,
        help="Plot trees with collapsing and coloring on amino acid level.",
    )
    parser_sim.set_defaults(func=simulate)

    # shared parameters
    for subparser in [parser_test, parser_infer, parser_sim]:
        subparser.add_argument(
            "--outbase", type=str, default="gctree.out", help="output file base name"
        )
        subparser.add_argument(
            "--img_type", type=str, default="svg", help="output image file type"
        )
        subparser.add_argument(
            "--verbose", action="store_true", help="flag for verbose messaging"
        )

    # common parameters for the inference and simulation subprograms
    for subparser in [parser_infer, parser_sim]:
        subparser.add_argument(
            "--frame", type=int, default=None, choices=(1, 2, 3), help="codon frame"
        )
        subparser.add_argument(
            "--idlabel",
            action="store_true",
            help=r"label nodes with their sequence ids in output tree images, "
            "and write a fasta alignment mapping those sequence ids to sequences. "
            "This is the easiest way to access inferred ancestral sequences.",
        )

    return parser


def main(arg_list=None):
    args = get_parser().parse_args(arg_list)
    args.func(args)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio import AlignIO
from Bio.Phylo.TreeConstruction import MultipleSeqAlignment
from collections import defaultdict, Counter


def fasta_parse(aln_file, root, frame=None, aln_file2=None, id_abundances=False):
    if aln_file.endswith("fasta") or aln_file.endswith("fa"):
        aln_format = "fasta"
    elif aln_file.endswith("phylip") or aln_file.endswith("phy"):
        aln_format = "phylip"
    else:
        raise ValueError("unrecognized alignment file type: " + aln_file)
    aln = AlignIO.read(aln_file, aln_format)
    if aln_file2 is not None:
        assert frame is None
        aln_combined = MultipleSeqAlignment([])
        if aln_file2.endswith("fasta") or aln_file2.endswith("fa"):
            aln_format2 = "fasta"
        elif aln_file2.endswith("phylip") or aln_file2.endswith("phy"):
            aln_format2 = "phylip"
        else:
            raise ValueError("unrecognized alignment file type: " + aln_file2)
        aln2 = AlignIO.read(aln_file2, aln_format2)
        for seq in aln:
            cell = seq.id[:-1] if seq.id != root else root
            for seq2 in aln2:
                cell2 = seq2.id[:-1] if seq2.id != root else root
                if cell2 == cell:
                    aln_combined.append(
                        SeqRecord(Seq(str(seq.seq) + str(seq2.seq)), id=cell)
                    )
        aln = aln_combined

    sequence_length = aln.get_alignment_length()
    if frame is not None:
        start = frame - 1
        end = start + 3 * ((sequence_length - start) // 3)
    else:
        start = 0
        end = sequence_length

    seqs_unique_counts = defaultdict(list)
    id_set = set()
    root_seq = None
    for seq in aln:
        # if id is just an integer, assume it represents count of that sequence
        if seq.id in id_set:
            raise ValueError("Sequence ID found multiple times:", seq.id)
        else:
            id_set.add(seq.id)
        seqstr = str(seq.seq)[start:end].upper()
        if seq.id == root:
            root_seq = seqstr
            if seqstr not in seqs_unique_counts:
                seqs_unique_counts[
                    seqstr
                ] = []  # no observed root unless we see it elsewhere
        elif seq.id.isdigit() and id_abundances:
            seqs_unique_counts[seqstr] = [seq.id for _ in range(int(seq.id))]
        else:
            seqs_unique_counts[seqstr].append(seq.id)

    if root_seq is None:
        raise ValueError("root seq id {} not found".format(root))

    new_aln = MultipleSeqAlignment([SeqRecord(Seq(root_seq), id=root)])
    counts = {
        root: len(seqs_unique_counts[root_seq])
    }  # Add the count for the root sequence
    id_map = {root: [x for x in seqs_unique_counts[root_seq] if x != root]}
    del seqs_unique_counts[root_seq]  # Now delete the root so it does not appear twice
    for i, seq in enumerate(seqs_unique_counts, 1):
        new_id = "seq" + str(i)
        new_aln.append(SeqRecord(Seq(seq), id=new_id))
        counts[new_id] = len(seqs_unique_counts[seq])
        id_map[new_id] = seqs_unique_counts[seq]

    return new_aln, counts, id_map


def check_header(header):
    try:
        header.decode("ascii")
    except UnicodeDecodeError as e:
        print("Sequence header must be an ascii-encoded string:", header)
        raise e
    if len(header) > 10:
        print("Sequence headers must be shorter than 10 characters:", header)
        raise Exception
    try:
        int(header)
        raise Exception(
            "Sequence headers must be distinguishable from an integer. Please add a non number character."
        )
    except ValueError:
        pass


def get_parser():
    parser = argparse.ArgumentParser(
        description=(
            "Deduplicate sequences in a fasta file, write to stdout in phylip format,"
            " and create a few other files (see arguments). Headers must be a unique"
            " ID of less than or equal to 10 ASCII characters. An additional sequence"
            " representing the outgroup/root must be included"
            " (even if one or more observed sequences are identical to it)."
        )
    )
    parser.add_argument(
        "infile",
        type=str,
        nargs="+",
        help="Fasta file with less than or equal to 10 characters unique header ID. "
        "Because dnapars will name internal nodes by integers a node name must include"
        "at least one non number character (unless using the id_abundances option).",
    )
    parser.add_argument(
        "--abundance_file",
        type=str,
        default=None,
        help="filename for the output file containing the counts.",
    )
    parser.add_argument(
        "--idmapfile",
        type=str,
        default=None,
        help="filename for the output file containing the map of new unique ids to original seq ids.",
    )
    parser.add_argument(
        "--id_abundances",
        action="store_true",
        help="flag to interpret integer ids in input as abundances",
    )
    parser.add_argument(
        "--root",
        type=str,
        default="root",
        help=(
            "ID of the root sequence in fasta file. This ID will be"
            " used as the unique id for the root sequence, and any"
            " observed sequences matching the root sequence."
        ),
    )
    parser.add_argument(
        "--frame", type=int, default=None, help="codon frame", choices=(1, 2, 3)
    )
    parser.add_argument(
        "--colorfile",
        type=str,
        default=None,
        help="optional input csv filename for colors of each cell.",
    )
    parser.add_argument(
        "--colormap",
        type=str,
        default=None,
        help="optional output filename for colors map.",
    )
    return parser


def main(arg_list=None):
    args = get_parser().parse_args(arg_list)

    new_aln, counts, id_map = fasta_parse(
        args.infile[0],
        args.root,
        frame=args.frame,
        aln_file2=args.infile[1] if len(args.infile) == 2 else None,
        id_abundances=args.id_abundances,
    )
    print(format(new_aln, "phylip"))
    if args.abundance_file is not None:
        with open(args.abundance_file, "w") as f:
            for seqID, count in counts.items():
                print("{},{}".format(seqID, count), file=f)
    if args.idmapfile is not None:
        with open(args.idmapfile, "w") as f:
            for seq_id, cell_ids in id_map.items():
                print("{},{}".format(seq_id, ":".join(cell_ids)), file=f)
    if args.colorfile is not None:
        with open(args.colorfile, "r") as f:
            cell_colors = dict(line.rstrip().split(",") for line in f)
        with open(args.colormap, "w") as f:
            for seq_id, cell_ids in id_map.items():
                colors_counts = list(
                    Counter(
                        cell_colors[cell_id] if cell_id in cell_colors else None
                        for cell_id in cell_ids
                    ).items()
                )
                if (
                    len(colors_counts) == 1
                    and colors_counts[0][0] is not None
                    and colors_counts != []
                ):
                    colors_str = colors_counts[0][0]
                    print("{}\t{}".format(seq_id, colors_str), file=f)
                elif len(colors_counts) > 1:
                    colors_strs = [
                        "{}:{}".format(color, count) for color, count in colors_counts
                    ]
                    print("{}\t{}".format(seq_id, ",".join(colors_strs)), file=f)



import setuptools
import versioneer


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gctree",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="William DeWitt",
    author_email="wsdewitt@gmail.com",
    description="phylogenetic inference of genotype-collapsed trees",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/matsengrp/gctree",
    entry_points={"console_scripts": ["gctree=gctree.cli:main",
                                      "deduplicate=gctree.deduplicate:main",
                                      "mkconfig=gctree.mkconfig:main",
                                      "isotype=gctree.isotype:main",
                                      "phylip_parse=gctree.phylip_parse:main"]},
    packages=['gctree'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7, <3.10",
    install_requires=[
        "PyQt5",
        "ete3",
        "biopython",
        "matplotlib",
        "pandas",
        "scipy",
        "seaborn",
        "multiset",
        "frozendict",
        "historydag"
    ],
)
